import numpy as np
from collections import namedtuple
from matplotlib import cm as cm, pyplot as plt, gridspec as gridspec
from matplotlib.patches import ConnectionPatch
from navigation_vis.AbstractView import AbstractView


class GridSpec(AbstractView):
    MAX_PHI_DIM = 3
    CellParams = namedtuple("CellParams", "vmin vmax xlim ylim xticks yticks spines", 
                            defaults=[None, None, None, None, [], [], False])
    GridParams = namedtuple("GridParams", "xticks yticks spines", 
                            defaults=[None, None, True])
    def __init__(self, grid_data):
        """
        :params grid_values: array of size (H x W x D), where D is number of features

        """
        self.grid_data = None
        self.H, self.W = None, None
        self.X, self.Y = None, None
        self.feature_shape = None
        self.fig = None
        self.gs = None
        self.axes_grid = None
        self.update_data(grid_data)
        
    def update_data(self, grid_data):
        self.grid_data = grid_data
        self.H, self.W = self.grid_data.shape[:2]
        self.X, self.Y = self.W + 1, self.H + 1
        self.feature_shape = self.grid_data.shape[2:]
        self.fig = plt.figure(figsize=(min(self.W * 2, 20), min(self.H * 2, 20)))
        self.gs = gridspec.GridSpec(self.H, self.W)
        self.gs.update(wspace=0., hspace=0., left=0., right=1., bottom=0., top=1.)
        self.axes_grid = {
            self._rowcol_to_xy(row, col): plt.Subplot(self.fig, self.gs[row, col]) \
            for row in range(self.H) \
            for col in range(self.W) \
            }
        
    def get_axis_xy(self, x, y):
        return self.axes_grid[(x, y)]

    def set_axis_xy(self, x, y, ax):
        self.axes_grid[(x, y)] = ax

    def get_axis_rc(self, r, c):
        return self.axes_grid[self._rowcol_to_xy(r, c)]

    def set_axis_rc(self, r, c, ax):
        self.axes_grid[self._rowcol_to_xy(r, c)] = ax

    def get_cell_center(self):

        n_dim = len(self.feature_shape)
        if n_dim == 0 or (n_dim == 1 and self.feature_shape[0] == 1):
            return 0, 0
        elif n_dim == 1 and self.feature_shape[0] > 1:
            return self.feature_shape[0] / 2 - 0.5, 0.5
        elif 2 <= n_dim <= 3:
            return (self.feature_shape[0] / 2 - (1 - self.feature_shape[0] % 2) * 0.5,  # -0.5 when shape is even
                    self.feature_shape[1] / 2 - (1 - self.feature_shape[1] % 2) * 0.5)
        else:
            return 0, 0

    def _prepare_cell_axis(self, row, col, cell_params):
        """Prepares gridspec axes.

        """
        ax = self.get_axis_rc(row, col)

        if cell_params is None:
            cell_params = self.CellParams()
        for sp in ax.spines.values():
            sp.set_visible(cell_params.spines)
        if cell_params.xlim is not None:
            ax.set_xlim(cell_params.xlim)
        if cell_params.ylim is not None:
            ax.set_ylim(cell_params.ylim)
        if cell_params.xticks is not None:
            ax.set_xticks(cell_params.xticks)
        if cell_params.yticks is not None:
            ax.set_yticks(cell_params.yticks)
        self.set_axis_rc(row, col, ax)
        return ax, cell_params

    def _prepare_grid_axis(self, row, col, grid_params):
        """Prepares gridspec axes.

        """
        ax = self.get_axis_rc(row, col)
        if grid_params is None:
            grid_params = self.GridParams()

        if ax.is_first_row():
            ax.spines['top'].set_visible(grid_params.spines)
        if ax.is_last_row():
            ax.spines['bottom'].set_visible(grid_params.spines)
            if grid_params.xticks is not None:
                ax.set_xticks(grid_params.xticks)
        if ax.is_first_col():
            ax.spines['left'].set_visible(grid_params.spines)
            if grid_params.yticks is not None:
                ax.set_yticks(grid_params.yticks)
        if ax.is_last_col():
            ax.spines['right'].set_visible(grid_params.spines)
        self.set_axis_rc(row, col, ax)
        return ax, grid_params

    def _connect_axes(self, ax1, ax2):
        """Draws an arrow from ax1 to ax2.

        """
        axis_center = self.get_cell_center()
        con = ConnectionPatch(xyA=axis_center, xyB=axis_center,
                              coordsA="data", coordsB="data",
                              axesA=ax1, axesB=ax2, color="red",
                              mutation_scale=40, arrowstyle="->",
                              shrinkB=5, shrinkA=5)
        ax1.add_artist(con)
        con = ConnectionPatch(xyA=axis_center, xyB=axis_center,
                              coordsA="data", coordsB="data",
                              axesA=ax2, axesB=ax1, color="red",
                              mutation_scale=40, arrowstyle="<-",
                              shrinkB=5, shrinkA=5)
        ax2.add_artist(con)
        ax1.plot(*axis_center, 'ro', markersize=10)
        ax2.plot(*axis_center, 'ro', markersize=10)

    def _xy_to_rowcol(self, x, y):
        """Converts (x, y) to (row, col).

        """
        return self.H - y, x - 1

    def _rowcol_to_xy(self, row, col):
        """Converts (row, col) to (x, y).

        """
        return col + 1, self.H - row

    def render_plot(self, row, col, cell_params, grid_params):
        """Render a cell of of the grid as subplot.

        """
        feature = self.grid_data[row, col]
        ax, cell_params = self._prepare_cell_axis(row, col, cell_params)
        ax, grid_params = self._prepare_grid_axis(row, col, grid_params)

        n_dim = len(feature.shape)
        if n_dim == 0 or (n_dim == 1 and feature.shape[0] == 1):
            ax.imshow(np.atleast_2d(feature), vmin=cell_params.vmin, vmax=cell_params.vmax)
        elif n_dim == 1 and feature.shape[0] > 1:
            ax.plot(feature)
        elif 1 < n_dim <= 3:
            ax.imshow(feature, vmin=cell_params.vmin, vmax=cell_params.vmax, cmap="gray" if n_dim == 2 else None)
        else:
            raise ValueError("Feature dimensions > {} not supported!".format(self.MAX_PHI_DIM))

        return ax

    def prepare_plots(self, cell_params=None, grid_params=None):
        for row in range(self.H):
            for col in range(self.W):
                self.render_plot(row, col, cell_params, grid_params)

    def render(self, cell_params=None, grid_params=None):
        """Render the grid
        """
        self.prepare_plots(cell_params, grid_params)
        for row in range(self.H):
            for col in range(self.W):
                self.fig.add_subplot(self.get_axis_rc(row, col))
        return self

    def add_trajectory(self, trajectory):
        """Adds a trajectory to the figure.

        :params trajectory: [(x1, y1), ...]
        """
        x_list, y_list = tuple(zip(*trajectory))  # [(x, y), ..] -> [x, ...], [y, ...]
        for idx in range(len(x_list) - 1):
            ax1 = self.axes_grid[(x_list[idx], y_list[idx])]
            ax2 = self.axes_grid[(x_list[idx + 1], y_list[idx + 1])]
            ax1.set_zorder(-2 * idx + 1)
            ax2.set_zorder(-2 * idx)
            self._connect_axes(ax1, ax2)

    def add_trajectories(self, traj_lst):
        """Adds a list of trajectories to the figure.

        :params traj_lst: list of trajectories, where trajectory = [(x1, y1), ...]
        """
        if traj_lst is not None:
            for traj in traj_lst:
                self.add_trajectory(traj)


import numpy as np
from matplotlib import cm as cm, pyplot as plt, colors as mplotcolors
from mpl_toolkits.axes_grid1 import make_axes_locatable
from navigation_vis.AbstractView import AbstractView

def image_grid_to_raster(image_grid):
    """ Converts image grid to RGB raster image.
    :param image_grid: numpy array of size (nY x nX x Y x X x C)
    :return: RGB image
    """
    """
    1. nY    x    nX    x    Y    x    X    x    C        -> swap axes
                   .       .
                     .   .
                       .
                     .   .
                   .       .
    2. nY    x    Y    x    nX    x    X    x    C        -> reshape merge axis 0 with 1, 2 with 3
    3. (nY   *   Y)    x    (nX   *   X)    x    C        = H x W x C
    """
    nY, nX, Y, X, C = image_grid.shape
    return image_grid.swapaxes(  # 1
        1, 2).reshape(  # 2
        nY * Y, nX * X, C)

def image_array_to_grid(image_array, empty_fill=0):
    """ Converts image array to grid array.

    :param image_array: numpy array of size (N x Y x X x C)
    :param empty_fill: what to fill empty blocks with
    :return: numpy array of size (nY x nX x Y x X x C)
    """
    # Ref: https://stackoverflow.com/questions/42040747/more-idiomatic-way-to-display-images-in-a-grid-with-numpy
    N, Y, X, C = image_array.shape
    #  N    x    Y     x    X    x    C        -> factor axis 1 into two axis
    # nY    x    nX    x    Y    x    X    x    C
    nX = int(np.ceil(np.sqrt(N)))
    nY = int(np.ceil(N / nX))  # nX >= nY
    n_empty_blocks = (nX * nY) - N
    if n_empty_blocks > 0:
        image_array = np.concatenate(
            [image_array,
             empty_fill * np.ones((n_empty_blocks, Y, X, C), dtype=image_array.dtype)
             ]
        )
    return image_array.reshape(nY, nX, Y, X, C)

def flatten_to_raster(data):
    """ Flatten numpy array of various dimensions to RGB raster image.
    :param data: numpy array of one of following sizes.
           1) H x W x C (color/gray image)
           2) N x Y x X x C (array of color/gray images)
           3) nY x nX x Y x X x C (2d array of color/gray images)
           (C has to be 1 or 3)
           E.g., C = 1
             <---------------W--------------->
              ------- ------- ------- -------
         ^   |       |       |       |       |
         |   Y   1   |   2   |  ...  |   nX  |
         |   |       |       |       |       |
         |    ---X--- ------- ------- -------
         |   |       |       |       |       |
         |   |   2   |       |       |       |
         |   |       |       |       |       |
         H    ------- ------- ------- -------
         |   |       |       |       |       |
         |   |  ...  |       |       |       |
         |   |       |       |       |       |
         |    ------- ------- ------- -------
         |   |       |       |       |       |
         |   |  nY   |       |       |       |
         v   |       |       |       |       |
              ------- ------- ------- -------
       Input -> Output
       1) H x W x C -> H x W x C
       2) N x Y x X x C -> (nY*Y) x (nX*X) x C
           where nX & nY are factors of N such that we get as close to square grid as possible
           (with bias towards having more columns than rows so for 12 images we have nY x nX = 3 x 4 grid)
       3) nH x nW x H x W x C -> (nY*Y) x (nX*X) x C
    """
    n_dim = len(data.shape)
    if n_dim == 3:
        H, W, C = data.shape
        nY, nX = 1, 1
        flattened = data
        n_states = nY * nX
    elif n_dim == 4:
        # N x Y x X x C (array of color/gray images)
        image_grid = image_array_to_grid(data)
        nY, nX, Y, X, C = image_grid.shape
        flattened = image_grid_to_raster(image_grid)  # .transpose(1,0,2,3,4)
        n_states = len(data)
    elif n_dim == 5:
        image_grid = data
        nY, nX, Y, X, C = image_grid.shape
        flattened = image_grid_to_raster(image_grid)
        n_states = nY * nX
    else:
        raise Exception("data dimension {} not supported!".format(n_dim))

    return flattened, nY, nX, n_states

def get_discrete_cmap(cmap, discrete_levels):

    n_unique = len(np.unique(discrete_levels))
    norm = mplotcolors.Normalize(vmin=0, vmax=len(discrete_levels))
    # Leave string colors as it is, convert int colors to
    # normalized rgba
    cell_colors = [cmap(norm(val)) for val in discrete_levels]
    cmap = mplotcolors.ListedColormap(cell_colors, N=n_unique)
    return cmap

def get_css4_colors(N, shuffled=False):

    colors = list(mplotcolors.CSS4_COLORS.keys())

    if shuffled:
        np.random.shuffle(colors)

    if N == 0:
        return None
    elif N < 0:
        return colors

    times = int(np.ceil(N / len(colors)))

    if times == 1:
        return colors[:N]
    else:
        colors_tiled = colors * times
        return colors_tiled[:N]

class Raster(AbstractView):

    def __init__(self, data, ax=None):
        if ax is None:
            self.fig, self.ax = plt.subplots(1, 1)
        else:
            self.ax = ax
        self.update_data(data)

    def update_data(self, data):
        self.data = data
        self.raster, self.nY, self.nX, self.n_states = flatten_to_raster(data)
        self.H, self.W, self.C = self.raster.shape
        self.Y, self.X = self.H // self.nY, self.W // self.nX  # guaranteed to be ints
        self.render_img = self.raster[:, :, 0] if self.raster.shape[2] == 1 else self.raster
        # self.render_img = self.render_img[::-1]
        # self.render_img = np.roll(self.render_img, -1, axis=0)

    def _cell_coord_to_cell_idx(self, xi, yi):
        return yi * self.nX + xi

    def _xi_to_x(self, xi, center=True):
        x = xi * self.X + (self.X // 2 if center else 0)
        assert x >= 0 and x < self.W
        return x

    def _yi_to_y(self, yi, center=True):
        y = yi * self.Y + (self.Y // 2 if center else 0)
        assert y >= 0 and y < self.H
        return y

    def add_pixel_trajectory(self, trajectory, with_arrow=True, arrow_props=dict(), color='white'):
        x_list, y_list, a_list = [], [], []
        for (x, y, a) in trajectory:
            x_list.append(x)
            y_list.append(y)
        if with_arrow:
            self.draw_path_with_arrows(x_list, y_list, color, arrow_props)
        else:
            self.draw_path(x_list, y_list, color)
        return self

    def add_trajectory(self, trajectory, with_arrow=True, arrow_props=dict(), color='white'):
        x_list, y_list, a_list = [], [], []
        for (x, y, a) in trajectory:
            x_list.append(x)
            y_list.append(y)
        x_list = [self._xi_to_x(xi) for xi in x_list]
        y_list = [self._yi_to_y(yi) for yi in y_list]
        if with_arrow:
            self.draw_path_with_arrows(x_list, y_list, color, arrow_props)
        else:
            self.draw_path(x_list, y_list, color)
        return self

    def add_pixel_trajectories(self, trajectories, with_arrow=True, arrow_props=dict(), color='white'):
        traj_color_list = get_css4_colors(len(trajectories), True)
        for i, traj in enumerate(trajectories):
            self.add_pixel_trajectory(traj, with_arrow, arrow_props, traj_color_list[i])
        return self

    def add_trajectories(self, trajectories, with_arrow=True, arrow_props=dict(), color='white'):
        traj_color_list = get_css4_colors(len(trajectories), True)
        for i, traj in enumerate(trajectories):
            self.add_trajectory(traj, with_arrow, arrow_props, traj_color_list[i])
        return self

    def render(self, cmap=cm.viridis):
        self.cmap = cmap
        self.im = self.ax.imshow(self.render_img, cmap=cmap)
        # self.ax.invert_yaxis()
        return self

    def get_raster_entry(self, x, y):
        return self.render_img[y, x]

    def draw_path(self, x_list, y_list, color):
        return plt.plot(x_list, y_list, color=color)

    def draw_path_with_arrows(self, x_list, y_list, color, arrow_props):
        if len(x_list) >= 2:
            for i in range(len(x_list)-1):
                x, y = x_list[i], y_list[i]
                xp, yp = x_list[i+1], y_list[i+1]
                self.ax.annotate("", xy=(xp, yp), xytext=(x, y), arrowprops={"color":color, "arrowstyle": "->", **arrow_props})
        return self

    def ticks(self, xticks=True, yticks=True, minor=True):
        show_minor_ticks = False
        if self.nX == 1 and self.nY == 1:
            # one image
            xtick_locs = np.arange(0, self.W, 1) - 0.5
            ytick_locs = np.arange(0, self.H, 1) - 0.5
        else:
            # image grid
            # xtick_locs = np.arange(0, self.W, self.X) + self.X // 2 - (0.5 if self.X % 2 == 0 else 0)
            # ytick_locs = np.arange(0, self.H, self.Y) + self.Y // 2 - (0.5 if self.Y % 2 == 0 else 0)
            xtick_locs = np.arange(0, self.W, self.X) - 0.5
            ytick_locs = np.arange(0, self.H, self.Y) - 0.5
            xminor_tick_locs = np.arange(0, self.W, 1)
            yminor_tick_locs = np.arange(0, self.H, 1)
            show_minor_ticks = True

        if xticks:
            # set major ticks
            self.ax.set_xticks(xtick_locs, minor=False)
            self.ax.set_xticklabels([str(tick_idx) for tick_idx, tick_val in enumerate(xtick_locs)], minor=False)
            # set minor ticks
            if minor and show_minor_ticks:
                self.ax.set_xticks(xminor_tick_locs, minor=True)

        if yticks:
            # set major ticks
            self.ax.set_yticks(ytick_locs, minor=False)
            self.ax.set_yticklabels([str(tick_idx) for tick_idx, tick_val in enumerate(ytick_locs)], minor=False)
            # set minor ticks
            if minor and show_minor_ticks:
                self.ax.set_yticks(yminor_tick_locs, minor=True)

        # tick properties
        self.ax.tick_params(which='both', width=1)
        self.ax.tick_params(which='major', length=7, color='red')
        self.ax.tick_params(which='minor', length=2, color='black')
        return self

    def grid(self, major=True, minor=False):
        if major:
            self.ax.grid(which='major', linestyle='-', linewidth='1', color='red')
        if minor:
            self.ax.grid(which='minor', linestyle=':', linewidth='0.3', color='black')
        return self

    def title(self, title):
        self.ax.set_title(title)
        return self

    def colorbar(self, ticks, ticklabels=None, fig=None):
        if fig is None:
            fig = self.fig
        if ticklabels is None:
            ticklabels = ticks
        divider = make_axes_locatable(self.ax)
        self.cb_ax = divider.append_axes('right', size='5%', pad=0.05)
        self.cbar = fig.colorbar(self.im, ticks=ticks, cax=self.cb_ax)
        self.cbar.set_ticklabels(ticklabels)
        return self

    def show_raster_text(self, fmt=".1f", color_cb=lambda elem: "white", fontsize=None):
        # Loop over data dimensions and add text annotations.
        for y in range(self.H):
            for x in range(self.W):
                if self.C == 1:
                    text =  format(self.get_raster_entry(x, y), fmt)
                else:
                    text = ", ".join([format(n, fmt) for n in self.get_raster_entry(x, y)])
                color = color_cb(self.get_raster_entry(x, y))
                # ax.text doesn't understand y axis inversion so do it for it
                self.ax.text(x, y, text, ha="center", va="center", color=color, fontsize=fontsize)
        return self

    def show_cell_text(self, text_lst=None, fmt=".1f", color_cb=lambda elem: "white", fontsize=None):
        # Loop over cells and add text annotations.
        for yi in range(self.nY):
            for xi in range(self.nX):
                x, y = self._xi_to_x(xi), self._yi_to_y(yi)
                state_idx = self._cell_coord_to_cell_idx(xi, yi)
                if state_idx >= self.n_states:
                    continue
                if text_lst is None:
                    if self.C == 1:
                        text =  format(self.get_raster_entry(x, y), fmt)
                    else:
                        text = ", ".join([format(n, fmt) for n in self.get_raster_entry(x, y)])
                    color = color_cb(self.get_raster_entry(x, y))
                else:
                    text = text_lst[state_idx]
                    color = color_cb(state_idx)
                # ax.text doesn't understand y axis inversion so do it for it
                self.ax.text(x, y, text, ha="center", va="center", color=color, fontsize=fontsize)
        return self

    def __call__(self, *args, **kwargs):
        return self.raster


import setuptools
# versioning style inspired from https://github.com/david-abel/simple_rl/
exec(open('navigation_vis/_version.py').read())

# Ref: https://packaging.python.org/tutorials/packaging-projects/

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="navigation_vis",
    version=__version__,
    author="Yagnesh Revar",
    author_email="mailto.yagnesh+github@gmail.com",
    description="A lightweight library for visualizing navigation grid world",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yrevar/navigation_vis",
    packages=setuptools.find_packages(),
    keywords = ['Markov Decision Process', 'MDP', 'Navigation', 'Data Visualization'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


class AbstractView:
    
    def update_data(self, data):
        raise NotImplementedError
        
    def add_trajectory(self, trajectory):
        raise NotImplementedError
        
    def add_trajectories(self, trajectories):
        raise NotImplementedError
    
    def render(self):
        raise NotImplementedError